--[[
	WindyLoads.lua
	
	Author: 	Sushifresser
	Date: 		20.04.2023
	Version: 	0.0.7.1
	
	Changelog:	V 0.0.7.1 @	20.04.2023	- bugfix with vehicles that may change configuration to not supported variants. Caused mod to be stuck on this vehicle
	Changelog:	V 0.0.7.0 @	18.04.2023	- moved tilt losses mechanic into separate mod "FS22_TiltLosses"
										- bugfix of crashing with particles on vehicles with fill units that have a capacity of 0
										- bugfix: vehicles with multiple fill units now shows particles for every fill unit
	Changelog:	V 0.0.6.1 @	29.03.2023	- bugfix in tilt loss procedure for non trailer vehicles
										- (trying to) bugfix in initialization of particle system on certain vehicles
	Changelog:	V 0.0.6.0 @	23.03.2023	- introduced tilting losses of vehicles
										- bugfix in particle system to prevent crashes on swap-body  vehicles
	Changelog:	V 0.0.5.1 @	20.01.2023	- bugfix in particle system initialization introduced by V0.0.5.0
	Changelog:	V 0.0.5.0 @	18.01.2023	- first implementation of particle system for blowing effect. See mod description for limitations
	Changelog:	V 0.0.4.1 @ 19.12.2022 	- bugfix of addSpecialization function in RegisterSpecialization.lua
	Changelog:	V 0.0.4.0 @ 07.02.2022 	- added some more description text
										- added possibility to determine the minimum required speed above which loads will be blown away by the modDesc.xml
										- added a global blow factor which is multiplied with the internal calculation results of the blow rates. It can be adjusted via the modDesc.xml. Standard is 1.0 which is equal to 100%. Setting this value to zero will effectively turn off the mod. Higher values increase the strength of the blwoing effect.
										- script now includes the possibility to exclude vehicles based on the user config from modDesc.xml
	Changelog:	V 0.0.3.0 @ 06.02.2022 	- first implementation of mod configuration via XML modDesc. Preparation for first publishing
	Changelog:	V 0.0.2.0 @ 10.12.2021 	- wind is now included in the calculation of the air speed above the vehicle
	Changelog:	V 0.0.1.0 @ 09.12.2021 	- initial implementation in FS 22
]]

WindyLoads = {};
local modName = g_currentModName;	--this global variable is only valid while the lua file is being loaded. So we need to store it in our own variable
local modDirectory = g_currentModDirectory;	--this global variable is only valid while the lua file is being loaded. So we need to store it in our own variable
WindyLoads.currentModDirectory = modDirectory;
source(modDirectory .. "WindyLoadsBlowEvent.lua");
WindyLoads.modDesc = loadXMLFile("modDesc", WindyLoads.currentModDirectory .. "modDesc.xml");
--some configuration options including their standard values
WindyLoads.MIN_SPEED_TO_BLOW = 20;	--the minimum speed below which no loads will be blown away
WindyLoads.DEBUG_MODE = false;	--determines whether debug information will be provided
WindyLoads.CONSIDER_WIND = true;	--determines whether the wind speed will be taken into account when calculating the relative air speed above the vehicle.
WindyLoads.BLOW_FACTOR_GLOBAL = 1;	--a global factor that can adjust the strength of the blow effect for all fill types
WindyLoads.ENABLE_PARTICLES = false;	--determines whether the paricles for the blow effect will be active
WindyLoads.EFFECT_FILE_LOCATION = "$data/effects/smoke/smokeParticles_3.i3d";	--file location of smoke particles definition from the game directory
WindyLoads.EMITTER_NODE_NAME = "smokeEmitter";	--the name of the smoke emitter node that we look for while initializing the particle system

WindyLoads.modConfigDone = false;	--marker for the completed initial configuration of this mod

WindyLoads.SYNC_TIME_INTERVAL_MS = 1000;	--the amount of milliseconds that is to pass between two synchronisation events of the fill unit levels

WindyLoads.EFFECT_FILENAME = WindyLoads.currentModDirectory .. "particles/smokeParticles.i3d"; --the perticel effect file that is delivered with the mod
WindyLoads.EMITTER_SHAPE_FILENAME = WindyLoads.currentModDirectory .. "particles/smokeEmitterShape.i3d"; --the particle emitter shape that is delivered with the mod

WindyLoads.PARTICLE_INTENSITY_FACTOR = 1.0;	-- a global factor to scale the particle intensity
WindyLoads.PARTICLE_MAX_INTENSITY = 0.35;	-- a global limit for the intensity of the particle effect

WindyLoads.vehicleID = 1;	--just some ID for counting on how many vehicle this mod has been initialized.

WindyLoads.hudData = {
--this determines how the hud will look if it used
	hudTextPosX = 0.99,
	hudTextPosY = 0.85,
	hudTextSize = 0.015 * g_gameSettings.uiScale
};
WindyLoads.applicableFillTypeCategories = {
--all FillTypeCategories, that are considered for this mod
	"LOADINGVEHICLE",
	"HAYLOFT",
	"WINDROW",
	"COMBINE",
	"FORK"
};

WindyLoads.excludedFillTypeCategories = {
--all FillTypeCategories which are to be excluded from this mod, even if they are defined in the applicable FillTypeCategories
	"LIQUID",
	"SPRAYER",
	"ANIMAL",
	"HORSE",
	"PRODUCT",
	"SLURRYTANK",
	"PIECE"
};

-- the standard weight per liter for unknown fill types
WindyLoads.STANDARD_MASS_PER_LITER = 0.0006;

-- the standard blow factor for unknown fill types
WindyLoads.STANDARD_BLOW_FACTOR = 0.5;

WindyLoads.blowFactors = {
--this table stores factors for each FillType that determines how strong the effect of the wind is.
--lower factors make the load to be blown away slower
	UNKNOWN = WindyLoads.STANDARD_BLOW_FACTOR,	--this is just to be sure
	WHEAT = 0.5,
	BARLEY = 0.5,
	OAT = 0.55,
	CANOLA = 0.5,
	SORGHUM = 0.5,
	SUNFLOWER = 0.6,
	SOYBEAN = 0.4,
	MAIZE = 0.3,
	SEEDS = 0.7,
	FORAGE = 0.5,
	CHAFF = 0.7,
	WOODCHIPS = 0.4,
	SILAGE = 0.15,
	GRASS_WINDROW = 0.8,
	DRYGRASS_WINDROW = 1,
	STRAW = 1,
	FERTILIZER = 0.15,
	MANURE = 0.05,
	LIME = 0.2
};

--this table holds the vent listeners that are used by the WindyLoads mod. We use it to add the event listeners and also remove them if necessary
WindyLoads.USED_EVENT_LISTENERS = {
		"onUpdate",	
		"onRootVehicleChanged",
		"onActivate",
		"onEnter",
		"onEnterVehicle",
		"onDraw" --only for debug hud so far
	};



--this is to store the relation between index and name of a FillType. Can be omitted later to work with the values given by the managers of the game.
--key: (int) fill type index; value: (string) fill type name
WindyLoads.fillTypesIndexToName = {};

--this will store all FillTypes which are considered applicable for this mod
--key: (string) fill type name; value: (table) fillType
WindyLoads.applicableFillTypes = {};


--this is to store vehicle types that are to be excluded from the effects of this mod
--key: (string) full vehicle name; value (boolean) whether the vehicle is to be excluded
WindyLoads.vehicleTypesToExclude = {};


--this function is used to init all required mod data
--this function may be called on a vehicle that has been initialized
--this function may also be called on a vehicle that was already initialized. It will be reset and reinitialized then
function WindyLoads.initModData(vehicle)

	vehicle.spec_windyLoads = {};
	local spec = vehicle.spec_windyLoads;
	--this stores the last calculated vehicle speed
	spec.lastVehicleSpeed = 0;

	--this table stores all fillUnits of the vehicle, that are susceptible to the mod's effects.
	--key: fillUnitIndex; value: table fillUnit
	spec.supportedFillUnits = {};
	
	--this table stores all blow rates of the corresponding fillUnits
	--key: fillUnitIndex; value: float blowRate
	spec.blowRatePerFillUnit = {};
	
	--this table holds the last blown volume values per fill unit
	--key: fillUnitIndex; value: float volume blown away in last update tick
	spec.lastBlownVolume = {};
	
	--this variable stores the passed time since the last synchronisation of fill levels for each fill unit individually
	spec.timeSinceLastSync = {};
	
	--just to be sure, we save the reference to our vehicle here
	spec.vehicle = vehicle;
	spec.vehicleID = WindyLoads.vehicleID;	--store the current vehicle ID in the spec of this mod for the current vehicle
	WindyLoads.vehicleID = WindyLoads.vehicleID + 1;	--increase the vehicle ID for this mod after every init process
	
	if not (WindyLoads.modConfigDone == true) then
		--try to find all fillTypes in the game that this mod can be applied to. They will be stored in the mod's class.
		WindyLoads.findApplicableFillTypes();
		--try to read the XML config to configure this mod with user settings
		WindyLoads:getXmlConfig()
		--now we're done
		WindyLoads.modConfigDone = true;
		
		if WindyLoads.DEBUG_MODE then
			print(string.format("[WindyLoads]: Mod successfully initialized with the following parameters: considerWindSpeed: %s; enableParticles: %s", tostring(WindyLoads.CONSIDER_WIND), tostring(WindyLoads.ENABLE_PARTICLES)));
		end;
			
		delete(WindyLoads.modDesc);
	end;
	
	--a table to hold several debug information
	spec.debugInfo = {};
	spec.debugInfo.vehicleWorldRotationX = 0;
	spec.debugInfo.vehicleWorldRotationY = 0;
	spec.debugInfo.vehicleWorldRotationZ = 0;
	
	
	spec.modInitialized = true;
	if WindyLoads.DEBUG_MODE then
		print(string.format("[WindyLoads]: Mod successfully loaded for vehicle '%s.'", Vehicle.getFullName(vehicle) .. WindyLoads.getVehicleIDStringDebug(vehicle)));
	end;

end;

function WindyLoads.prerequisitesPresent(specializations)
	local applicable = false;
	applicable = SpecializationUtil.hasSpecialization(FillUnit, specializations) and SpecializationUtil.hasSpecialization(FillVolume, specializations);
	applicable = applicable and SpecializationUtil.hasSpecialization(Dischargeable, specializations);
	return applicable;
end;

function WindyLoads.registerEventListeners(vehicleType)
	for _, functionName in ipairs(WindyLoads.USED_EVENT_LISTENERS) do
		SpecializationUtil.registerEventListener(vehicleType, functionName, WindyLoads);	--for getting all required info
		--print(string.format("registered eventlistener %s for vehicle Type %s", functionName, vehicleType.name));
	end;
	--the following event listeners are kept also when we decide to remove the other event listeners if a vehicle is not considered for the mod anymore
	--SpecializationUtil.registerEventListener(vehicleType, "onLoad", WindyLoads);
	SpecializationUtil.registerEventListener(vehicleType, "onLoadFinished", WindyLoads);
	SpecializationUtil.registerEventListener(vehicleType, "onDelete", WindyLoads);	--for removing some stuff
	SpecializationUtil.registerEventListener(vehicleType, "onVehicleSettingChanged", WindyLoads);	--for updating if the vehicle config is changed
end;

--this function is intended to remove a given vehicle from all the event listeners.
--This can be used for example if we found out that the vehicle cannot be considered for the mod any further
function WindyLoads.removeFromEventListeners(vehicleType)
	for _, functionName in ipairs(WindyLoads.USED_EVENT_LISTENERS) do
		--SpecializationUtil.removeEventListener(WindyLoads, functionName, WindyLoads);
		SpecializationUtil.removeEventListener(vehicleType, functionName, WindyLoads);
		if WindyLoads.DEBUG_MODE then
			print(string.format("[WindyLoads]: removed event listener %s for vehicle Type %s", functionName, vehicleType.name));
		end;
	end;
end;

--just some helper debug function to create better debug string outputs including the mods vehicle id
function WindyLoads.getVehicleIDStringDebug(vehicle)
	local result = "";
	if vehicle ~= nil then
		local spec = vehicle.spec_windyLoads;
		if spec ~= nil then
			local id = spec.vehicleID;
			if id ~= nil then
				result = "; [ID]: " .. tostring(id);
			end;
		end;
	end;
	return result;
end;

function WindyLoads.registerFunctions(vehicleType)
	--SpecializationUtil.registerFunction(vehicleType, "getVehicleIsSupportedByWindyLoads", WindyLoads.getVehicleIsSupportedByWindyLoads);
	SpecializationUtil.registerFunction(vehicleType, "getVehicleIsSupportedByWindyLoads", WindyLoads.getVehicleIsSupportedByWindyLoads);
	SpecializationUtil.registerFunction(vehicleType, "blowFromFillUnit", WindyLoads.blowFromFillUnit);
end;

--this function is called on each vehicle (self) a while loading the savegame
function WindyLoads:onLoadFinished(savegame)
	if WindyLoads.DEBUG_MODE then
		print(string.format("[WindyLoads]: Event 'onLoadFinished' on vehicle '%s.'", Vehicle.getFullName(self) .. WindyLoads.getVehicleIDStringDebug(self)));
	end;
	WindyLoads.initModData(self);	--call the initialization routine on this vehicle
end;

--this function is called on each vehicle (self) when the vehicle is deleted
function WindyLoads:onDelete()
	if WindyLoads.DEBUG_MODE then
		print(string.format("[WindyLoads]: Event 'onDelete' on vehicle '%s.'", Vehicle.getFullName(self) .. WindyLoads.getVehicleIDStringDebug(self)));
	end;
	WindyLoads:removeParticleSystems();
	self.spec_windyLoads = nil;
	if WindyLoads.DEBUG_MODE then
		print(string.format("[WindyLoads]: Event 'onDelete' on vehicle '%s.' finished", Vehicle.getFullName(self) .. WindyLoads.getVehicleIDStringDebug(self)));
	end;
end;

--this function is called on each vehicle (self) a while loading the savegame
function WindyLoads:onLoad(savegame)
	if WindyLoads.DEBUG_MODE then
		print(string.format("[WindyLoads]: Event 'onLoad' on vehicle '%s.'", Vehicle.getFullName(self) .. WindyLoads.getVehicleIDStringDebug(self)));
	end;
	WindyLoads:removeParticleSystems();
	WindyLoads.initModData(self);	--call the initialization routine on this vehicle
end;

--this function is called when the vehicle configuration is cahnged
function WindyLoads:onVehicleSettingChanged(gameSettingId, state)
	if WindyLoads.DEBUG_MODE then
		print(string.format("[WindyLoads]: Event 'onVehicleSettingChanged' on vehicle '%s.'", Vehicle.getFullName(self) .. WindyLoads.getVehicleIDStringDebug(self)));
	end;
	WindyLoads:removeParticleSystems();
	--WindyLoads.initModData(self);	--call the initialization routine on this vehicle
	WindyLoads:checkInitParticleSystem(self);
end;

--this function is called when the vehicle is entered
function WindyLoads:onEnterVehicle()
	if WindyLoads.DEBUG_MODE then
		print(string.format("[WindyLoads]: Event 'onEnterVehicle' on vehicle '%s.'", Vehicle.getFullName(self) .. WindyLoads.getVehicleIDStringDebug(self)));
	end;
	WindyLoads.checkInitParticleSystem(self, self);
end;

--this function is called when the root vehicle is changed
function WindyLoads:onRootVehicleChanged()
	if WindyLoads.DEBUG_MODE then
		print(string.format("[WindyLoads]: Event 'onRootVehicleChanged' on vehicle '%s.'", Vehicle.getFullName(self) .. WindyLoads.getVehicleIDStringDebug(self)));
	end;
	WindyLoads:removeParticleSystems();
	WindyLoads.checkInitParticleSystem(self, self);
end;

--this function is called when the vehicle is entered
function WindyLoads:onEnter()
	if WindyLoads.DEBUG_MODE then
		print(string.format("[WindyLoads]: Event 'onEnter' on vehicle '%s.'", Vehicle.getFullName(self) .. WindyLoads.getVehicleIDStringDebug(self)));
	end;
	WindyLoads.checkInitParticleSystem(self, self);
end;

--this function is called when the vehicle is activated
function WindyLoads:onActivate()
	if WindyLoads.DEBUG_MODE then
		print(string.format("[WindyLoads]: Event 'onActivate' on vehicle '%s.'", Vehicle.getFullName(self) .. WindyLoads.getVehicleIDStringDebug(self)));
	end;
	WindyLoads.checkInitParticleSystem(self, self);
end;

--this function checks if the particle system of the vehicle needs to be initialized
function WindyLoads:checkInitParticleSystem(vehicle)
    if WindyLoads.ENABLE_PARTICLES and vehicle.spec_windyLoads then
		local spec = vehicle.spec_windyLoads;
		if spec == nil or not spec.modInitialized then
			return;
		end;
		if spec.vehicleSupported then
			--here we need to init our particle system
			for fillUnitIndex, fillUnit in pairs(spec.supportedFillUnits) do
				if spec.windyParticleSpec and spec.windyParticleSpec[fillUnitIndex] and spec.windyParticleSpec[fillUnitIndex].particlesInitialized == true then
					--nothung to do. particles already successfully initialized
				else
					local particleInit = WindyLoads:getParticleSystems(vehicle, fillUnitIndex);
					--now check if everything has been set up correctly
					if not particleInit or not spec.windyParticleSpec[fillUnitIndex].particlesInitialized == true then
						if spec.windyParticleSpec then
							spec.windyParticleSpec[fillUnitIndex] = nil;
						else
							spec.windyParticleSpec = nil;
						end;
					end;
				end;
			end;
		end;
	end;
end;

--sometimes we should remove the existing particle systems from our vehicle
function WindyLoads:removeParticleSystems()
	local spec = self.spec_windyLoads;
	if not spec then
		return;
	end;
	if WindyLoads.ENABLE_PARTICLES and spec.windyParticleSpec then
		local specFillUnit = self.spec_fillUnit;
		if specFillUnit ~= nil then
			for fillUnitIndex, _ in pairs(specFillUnit:getFillUnits()) do
				local windyParticleSpec = spec.windyParticleSpec[fillUnitIndex];
				if windyParticleSpec ~= nil then
					ParticleUtil.deleteParticleSystem(windyParticleSpec.psEffects);
					unlink(windyParticleSpec.effect);
					unlink(windyParticleSpec.emitter);
					delete(windyParticleSpec.effect);
					delete(windyParticleSpec.emitter);
				end;
				--windyParticleSpec.particlesInitialized = nil;
				windyParticleSpec = nil;
			end;
		end;
	end;
end;


--this function tries to extract some configuration values from the mod description XML file.
function WindyLoads:getXmlConfig()

	local modDesc = WindyLoads.modDesc;
	
	WindyLoads.DEBUG_MODE = Utils.getNoNil(getXMLBool(modDesc, "modDesc.windyLoadsConfigurations#debugMode"), WindyLoads.DEBUG_MODE);	--we read the debug mode option from the modDesc file
	WindyLoads.CONSIDER_WIND = Utils.getNoNil(getXMLBool(modDesc, "modDesc.windyLoadsConfigurations#considerWindSpeed"), WindyLoads.CONSIDER_WIND);	--we read the consider wind speed option from the modDesc file
	WindyLoads.MIN_SPEED_TO_BLOW = Utils.getNoNil(getXMLFloat(modDesc, "modDesc.windyLoadsConfigurations#minBlowSpeed"), WindyLoads.MIN_SPEED_TO_BLOW);	--we read the minimum speed to start blow effect value from the modDesc file
	WindyLoads.BLOW_FACTOR_GLOBAL = Utils.getNoNil(getXMLFloat(modDesc, "modDesc.windyLoadsConfigurations#blowFactorGlobal"), WindyLoads.BLOW_FACTOR_GLOBAL);	--we read the globals scaling factor for the blow effect value from the modDesc file
	WindyLoads.BLOW_FACTOR_GLOBAL = math.max(WindyLoads.BLOW_FACTOR_GLOBAL, 0);	--we should not allow negative values as this would effectively add loads to the vehicle
	WindyLoads.ENABLE_PARTICLES = Utils.getNoNil(getXMLBool(modDesc, "modDesc.windyLoadsConfigurations#enableParticles"), WindyLoads.ENABLE_PARTICLES);	--we read the enable particles option from the modDesc file
	local specializationNumber = 0;
	--at first we take care of the custom fill type configurations
	while true do
		local specializationKey = string.format("modDesc.windyLoadsConfigurations.fillTypeConfigurations.fillTypeConfiguration(%d)", specializationNumber);
		if not hasXMLProperty(modDesc, specializationKey) then
			break;
		end;
		local fillTypeName = (Utils.getNoNil(getXMLString(modDesc, specializationKey .. "#fillTypeName"), "")):upper();
		if fillTypeName ~= "" then
			local fillTypeInclude = Utils.getNoNil(getXMLBool(modDesc, specializationKey .. "#fillTypeInclude"), false);
			local fillTypeExclude = Utils.getNoNil(getXMLBool(modDesc, specializationKey .. "#fillTypeExclude"), false);
			local fillTypeBlowFactor = Utils.getNoNil(getXMLFloat(modDesc, specializationKey .. "#fillTypeBlowFactor"), WindyLoads.STANDARD_BLOW_FACTOR);
			if fillTypeInclude and fillTypeExclude then
			--only one of the attributes should be provided. We will handle this item as to be excluded.
				Logging.xmlWarning(modDesc, "[WindyLoads]: WindyLoads recieved same fillType '%s' to be included an excluded at the same time. Item will be handled as to be excluded. Please check your configuration!", fillTypeName);
				fillTypeInclude = false;
			end;
			if fillTypeExclude then
				if WindyLoads.applicableFillTypes[fillTypeName] ~= nil then
					WindyLoads.fillTypesIndexToName[WindyLoads.applicableFillTypes[fillTypeName].index] = nil;
					WindyLoads.applicableFillTypes[fillTypeName] = nil;
					WindyLoads.blowFactors[fillTypeName] = nil;
					if WindyLoads.DEBUG_MODE then
						print(string.format("[WindyLoads]: Removed FillType '%s' from applicable FillTypes of WindyLoads mod based on user-defined XML configuration.", fillTypeName));
					end;
				end
			elseif fillTypeInclude then
				if WindyLoads.applicableFillTypes[fillTypeName] == nil then
				-- the fill type needs to be added at first
					local fillType = g_fillTypeManager:getFillTypeByName(fillTypeName);
					if fillType~= nil and fillType.name ~= nil then
						WindyLoads.applicableFillTypes[fillType.name] = fillType;
						WindyLoads.fillTypesIndexToName[fillType.index] = fillType.name;
						if WindyLoads.DEBUG_MODE then
							print(string.format("[WindyLoads]: Added FillType '%s' to applicable FillTypes of WindyLoads mod based on user-defined XML configuration.", fillType.name));
						end;
					end;
				end;
				if WindyLoads.applicableFillTypes[fillTypeName] == nil then
				-- we know the fill type already and now we can adjust its parameters
					WindyLoads.blowFactors[fillTypeName] = fillTypeBlowFactor;
				end;
			end;
		end;
			
		specializationNumber = specializationNumber + 1;
	end;
	specializationNumber = 0;
	-- now take another round for the custom vehicle type configurations
	while true do
		local specializationKey = string.format("modDesc.windyLoadsConfigurations.vehicleTypeConfigurations.vehicleTypeConfiguration(%d)", specializationNumber);
		if not hasXMLProperty(modDesc, specializationKey) then
			break;
		end;
		local vehicleTypeName = (Utils.getNoNil(getXMLString(modDesc, specializationKey .. "#fullVehicleName"), "")):upper();
		if vehicleTypeName ~= "" then
			local vehicleTypeExclude = Utils.getNoNil(getXMLBool(modDesc, specializationKey .. "#vehicleExclude"), false);
			if WindyLoads.vehicleTypesToExclude[vehicleTypeName] == nil then
				WindyLoads.vehicleTypesToExclude[vehicleTypeName] = true; -- now we remember the vehicle and can check at any time we have to
			end;
		end;
		
		specializationNumber = specializationNumber + 1;
	end;
end;

--to get the velocity of the current vehicle
function WindyLoads:getVehicleVelocity()
	local spec = self.spec_windyLoads;
	if not spec.modInitialized then
		return;
	end;
	local velocity = self:getLastSpeed();
	if velocity == nil or not velocity or velocity < 0 then
		return 0;
	end
	return velocity;
end;

function WindyLoads:findApplicableFillTypes()
--this function determines all applicable fillTypes based on the definitions of the FillTypeCategories
	local applicableFillTypes = {};
	local fillTypesToCheck = g_fillTypeManager:getFillTypes();
	Logging.devInfo("[WindyLoads]: Received " .. #fillTypesToCheck .. " from FillType Manager to check.");
	--WindyLoads:print_table(fillTypesToCheck, "fillTypesToCheck", "[WindyLoads]: The following fill types have been detected in the game: ");
	for _,fillType in pairs(fillTypesToCheck) do
		if WindyLoads.applicableFillTypes[fillType.name] == nil then
		--only go on if the FillType at hand is not already known
			local isApplicable = false;
			for _, applicableFillTypeCategory in pairs(WindyLoads.applicableFillTypeCategories) do
				if g_fillTypeManager:getIsFillTypeInCategory(fillType.index, applicableFillTypeCategory) then
					isApplicable = true;
					break;
				end;
			end;
			if isApplicable then
			--we found a FillType that seems to be applicable. Now we have to check if it is still to be excluded.
				for _, excludedFillTypeCategory in pairs(WindyLoads.excludedFillTypeCategories) do
					if g_fillTypeManager:getIsFillTypeInCategory(fillType.index, excludedFillTypeCategory) then
						isApplicable = false;
						break;
					end;
				end;
			end;
			if isApplicable then
			--the current fillType has passed all checks and can now be considered applicable
				WindyLoads.applicableFillTypes[fillType.name] = fillType;
				WindyLoads.fillTypesIndexToName[fillType.index] = fillType.name;	-- this table is for more comfortable access to the data
				if WindyLoads.DEBUG_MODE then
					print(string.format("[WindyLoads]: Added FillType '%s' to applicable FillTypes of WindyLoads mod based on standard configuration.", fillType.name));
				end;
			end;
		end;
	end;
	--WindyLoads:print_table(WindyLoads.applicableFillTypes, "WindyLoads.applicableFillTypes", "[WindyLoads]: The following fill types have been registered to be applicable for this mod: ");
end;

---Checks if the vehicle is valid for this mod to be applied to
-- @param table optional vehicle: The reference to the vehicle that is to be examined
-- @return boolean true if the vehicle is supported by this mod. Returns false otherwise
-- @return table Returns table of supported fillUnits of the vehicle if vehicle is supported. Returns nil otherwise.
function WindyLoads:getVehicleIsSupportedByWindyLoads(vehicle)
--this function should check if this vehicle is applicable for this mod
--TODO: maybe find some more useful criteria for checking the vehicles
	-- some vehicle types could be further excluded by default, e.g. combines, balers?
	--local self = vehicle or self;
	local spec = vehicle.spec_windyLoads;
	local specFillUnit = vehicle.spec_fillUnit;
	local vehicleSupported = false;
	
	if WindyLoads.vehicleTypesToExclude[Vehicle.getFullName(vehicle):upper()] ~= nil and WindyLoads.vehicleTypesToExclude[Vehicle.getFullName(vehicle):upper()] == true then
		--our list of vehicle types to excludes knows this vehicle type. So we can exit early
		if WindyLoads.DEBUG_MODE then
			print(string.format("[WindyLoads]: Vehicle '%s' is excluded for this mod. It will not be considered any further.", Vehicle.getFullName(vehicle) .. WindyLoads.getVehicleIDStringDebug(vehicle)));
		end;
		WindyLoads.removeFromEventListeners(g_vehicleTypeManager:getTypeByName(vehicle.typeName));
		return false;
	end;
	
	if specFillUnit == nil then	--we obviously need a fill unit in the vehicle for this mod
		if WindyLoads.DEBUG_MODE then
			print("[WindyLoads]: Could not find any FillUnit spec on this vehicle. Vehicle will not be considered for this mod any further");
		end;
		return false;
	end;
	local fillUnits = specFillUnit:getFillUnits();
	if fillUnits == nil or  not (#fillUnits > 0) then
		if WindyLoads.DEBUG_MODE then
			print("[WindyLoads]: Could not find any FillUnits on this vehicle. Vehicle will not be considered for this mod any further");
		end;
		return false;
	end;
	if WindyLoads.DEBUG_MODE then
		print(string.format("[WindyLoads]: Found " .. #fillUnits .. " FillUnits on vehicle '%s'.", Vehicle.getFullName(vehicle)));	
	end;
	
	local specFillVolume = vehicle.spec_fillVolume;	--we also need a fill volume for this mode. Otherwise it may not be an open vehicle which can be affected by wind
	if specFillVolume == nil then
		if WindyLoads.DEBUG_MODE then
			print("[WindyLoads]: Could not find any FillVolume spec on this vehicle. Vehicle will not be considered for this mod any further");
		end;
		return false;
	end;
	
	local specDischargeable = vehicle.spec_dischargeable;
	if specDischargeable == nil then
		if WindyLoads.DEBUG_MODE then
			print("[WindyLoads]: Could not find any Dischargeable spec on this vehicle. Vehicle will not be considered for this mod any further");
		end;
		return false;
	end;
	
	local supportedFillUnits = {};
	for fillUnitIndex, fillUnit in pairs(fillUnits) do
		local fillUnitSupported = false;
		local supportedFillTypes = specFillUnit:getFillUnitSupportedFillTypes(fillUnitIndex);
		if supportedFillTypes ~= nil then
			for fillTypeIndex, _ in pairs(supportedFillTypes) do
			--loop through all supported fillUnitFillTypes to find out if one is included which we can support with our mod
				for fillTypeName, fillType in pairs(WindyLoads.applicableFillTypes) do
					if fillType.index == fillTypeIndex then
						fillUnitSupported = true;
						break;
					end;
				end;
				if fillUnitSupported then
				-- we found the current fillUnit to be supported by our mod. Save it and then go on to the next fillUnit.
					supportedFillUnits[fillUnitIndex] = fillUnit;
					spec.blowRatePerFillUnit[fillUnitIndex] = 0;	--initialize the blow rates table
					spec.timeSinceLastSync[fillUnitIndex] = 0;		--initialize the sync times table
					spec.lastBlownVolume[fillUnitIndex] = 0;		--initialize the blow volumes table
					vehicleSupported = true;
					break;
				end;
				
			end;
		end;
		--if vehicleSupported then break end;
	end;
	
	if WindyLoads.DEBUG_MODE then
		print(string.format("[WindyLoads]: Vehicle '%s' was %s successfully checked for being applicable for WindyLoads mod.", Vehicle.getFullName(vehicle) .. WindyLoads.getVehicleIDStringDebug(vehicle), (vehicleSupported and "" or "not")));
	end;

	return vehicleSupported, supportedFillUnits;
end;

--this function is intended to extract the particle systems for a vehicle if we want to use it to show the blowing effect
function WindyLoads:getParticleSystems(vehicle, fillUnitIndex)
	--local self = vehicle or self;
	local spec = vehicle.spec_windyLoads;
	spec.windyParticleSpec = spec.windyParticleSpec or {};	--this will hold all info about the particle system of this vehicle
	if WindyLoads.ENABLE_PARTICLES ~= true then
		if WindyLoads.DEBUG_MODE then
			print("[WindyLoads]: Particle systems initialization was called but particles are not enabled. Aborting.");
			return false;
		end;
	end;
	local specFillUnit = vehicle.spec_fillUnit;
	local specDischargeable = vehicle.spec_dischargeable;
	if specDischargeable == nil then	--we want the dischargeable spec to be present because it gives a good indication that this vehicle is also allowed to being emptied. Otherwise we could deal with some kind of machine that can ust be filled and processes its load afterwards. We wouldn't want to affect such a machine with this mod
		if WindyLoads.DEBUG_MODE then
			print("[WindyLoads]: Couldn't find specialization 'Dischargeable' on vehicle " .. vehicle:getFullName() .. WindyLoads.getVehicleIDStringDebug(vehicle) .. ". Aborting!.");
			return false;
		end;
	end;
	
	--[[ in this table all required data about the particle system shall be stored. base elements:
		- effectEmitterNode: table; key: fill unit index; value: the node that is used for emitting the particles for the blow effect
		- psEffects: specific particle system that stores all the info about the particles to be emitted from the emitter node
	--]]
	spec.windyParticleSpec[fillUnitIndex] = {};	--an empty object to hold all info about the particle system of the current fill unit
	local windyParticleSpec = spec.windyParticleSpec[fillUnitIndex];	
	windyParticleSpec.particlesInitialized = false;	--just a flag for a successful init routine

	-- in this table all emitter nodes are stored sorted by their fill unit index.
	--key: fillUnitIndex; value: table the node for emitting particle effects
	--windyParticleSpec.effectEmitterNode = {};

	-- in this table all effects for the particle system shall be stored that belong to a fill unit.
	--key: fillUnitIndex; value: table, the particle system
	windyParticleSpec.psEffects = {};
	
	windyParticleSpec.effectNode = {};
	windyParticleSpec.emitterNode = {};
	
	local particleRootNode = nil;	--to this node we want to link the paricle emitter later
	local exactFillRootNode = specFillUnit:getFillUnitExactFillRootNode(fillUnitIndex); --the root fill node of the fill unit shall be our root for the emitter of the particles
	local autoAimNode = specFillUnit:getFillUnitAutoAimTargetNode(fillUnitIndex);	--we try to find the fill unit auto aim target node
	local fillUnitRootNode = specFillUnit:getFillUnitRootNode(fillUnitIndex);   --we try to get the base root node of the fill unit
    particleRootNode = autoAimNode;
	if particleRootNode == nil then	--find another root node if that went wrong
		--particleRootNode = specFillUnit:getFillUnitExactFillRootNode(fillUnitIndex); --the root fill node of the fill unit shall be our root for the emitter of the particles
		particleRootNode = exactFillRootNode;
	end;
    if particleRootNode == nil then	--find another root node if that went wrong
		particleRootNode = fillUnitRootNode;
	end;
	if WindyLoads.DEBUG_MODE then
			print(string.format("[WindyLoads]: Root nodes for fill unit %s on vehicle '%s': particleRootNode: %s; autoAimNode: %s; exactFillRootNode: %s; baseRootNode. %s", fillUnitIndex,  Vehicle.getFullName(vehicle) .. WindyLoads.getVehicleIDStringDebug(vehicle), particleRootNode, autoAimNode, exactFillRootNode, fillUnitRootNode));
		end;
	if particleRootNode ~= nil then
		windyParticleSpec.effectEmitterNode = particleRootNode;	--store the root node for the blow particle effect of this fill unit
	else --couldn't find root node for particle effect --> disable the particles for this fill unit
		windyParticleSpec.effectEmitterNode = nil;
		if WindyLoads.DEBUG_MODE then
			print(string.format("[WindyLoads]: Could not find particle root node for fill unit %s on vehicle '%s'.", fillUnitIndex,  Vehicle.getFullName(vehicle) .. WindyLoads.getVehicleIDStringDebug(vehicle)));
		end;
		return false;	--we couldn't find a root node for this fill unit. So we abort here and will not have the particle effect for this fill unit
	end;
	--find effect and store it in local spec
	local effect = g_i3DManager:loadSharedI3DFile(WindyLoads.EFFECT_FILENAME);
	local emitter = g_i3DManager:loadSharedI3DFile(WindyLoads.EMITTER_SHAPE_FILENAME);
	local linkNode = windyParticleSpec.effectEmitterNode;
	if effect == nil or emitter == nil then
        if WindyLoads.DEBUG_MODE then
			print(string.format("[WindyLoads]: Could not find particle effect files for fill unit %s on vehicle '%s'.", fillUnitIndex,  Vehicle.getFullName(vehicle) .. WindyLoads.getVehicleIDStringDebug(vehicle)));
		end;
		return false;	--we couldn't find the effect files for this fill unit. So we abort here and will not have the particle effect for this fill unit
    end;
    link(linkNode, effect);
	link(linkNode, emitter);
	windyParticleSpec.effect = effect;
	windyParticleSpec.emitter = emitter;
	windyParticleSpec.effectNode = getChildAt(effect, 0);	--store the node id of the effect node for later use
	windyParticleSpec.emitterNode = getChildAt(emitter, 0);	--store the node id of the emitter node for later use
	--setTranslation(emitter, 0, 0, 0);	--we could program some offset for or particle emitter
	local ps = {};
	ParticleUtil.loadParticleSystemFromNode(windyParticleSpec.effectNode, ps, false, true);
	ParticleUtil.setEmitterShape(ps, windyParticleSpec.emitterNode);
	windyParticleSpec.psEffects = ps;
	if not getHasShaderParameter(windyParticleSpec.psEffects.shape, "colorAlpha") then
		if WindyLoads.DEBUG_MODE then
			print(string.format("[WindyLoads]: Missing shader attribute 'colorAlpha' in WindyLoads particle system material for fill unit %s on vehicle '%s'. Could not set up particle system correctly.", fillUnitIndex,  Vehicle.getFullName(vehicle) .. WindyLoads.getVehicleIDStringDebug(vehicle)));
		end;
		return false;
	end
	windyParticleSpec.particlesInitialized = true;
	if WindyLoads.DEBUG_MODE then
		print("[WindyLoads]: Successfully added 'WindyLoads' particle system to fill unit " .. fillUnitIndex .. " of vehicle '" .. vehicle:getFullName() .. WindyLoads.getVehicleIDStringDebug(vehicle) .."'");
	end;
	return true;
end;

---This function calculates the current relative air speed above the vehicle. This is composed of the current vehicle velocity and direction above ground as well as the current wind speed and direction above ground.
-- @param table vehicle: The reference to the vehicle that is to be considered. Defaults to the own vehicle of this mod.
function WindyLoads:calculateCurrentRelativeAirSpeed(vehicle)
	local self = vehicle or self;
	local spec = vehicle.spec_windyLoads;
	local vehicleSpeed = vehicle:getLastSpeed();
	spec.lastVehicleSpeed = vehicleSpeed or 0;
	local airSpeed = 0;
	if WindyLoads.CONSIDER_WIND then
		local windDirX, windDirZ, windVelocity = g_currentMission.environment.weather.windUpdater:getCurrentValues();
		spec.debugInfo.windDirX = windDirX or 0;
		spec.debugInfo.windDirZ = windDirZ or 0;
		spec.debugInfo.windVelocity = windVelocity or 0;
		local vehicleWorldDirection = {};
		vehicleWorldDirection.X, vehicleWorldDirection.Y, vehicleWorldDirection.Z = vehicle:getVehicleWorldDirection();
		spec.debugInfo.vehicleWorldDirectionX = vehicleWorldDirection.X or 0;
		spec.debugInfo.vehicleWorldDirectionY = vehicleWorldDirection.Y or 0;
		spec.debugInfo.vehicleWorldDirectionZ = vehicleWorldDirection.Z or 0;
		
		--now calculate the effective relative air spedd above the vehicle
		airSpeed = math.sqrt(vehicleSpeed^2 + windVelocity^2 + 2*windVelocity*vehicleSpeed*(vehicleWorldDirection.X*-windDirX+vehicleWorldDirection.Z*-windDirZ));
	else
		airSpeed = vehicleSpeed;
	end;
	spec.lastAirSpeed = airSpeed;
	return airSpeed;
end;

--this function calculates a value which represents the rate at which the load of a vehicle fill unit is blown away by the wind.
-- @param float airSpeed: this parameter states the air speed above the fill unit that is used to compute the blow rates
-- @param table fillUnits: A table that contains the fill unit indices as keys and the fillUnits to be examined as values. Defaults to an emtpy table.
-- @return table: Returns a table of blowrates (float values) for each FillUnit of the vehicle. The value represents percent of the current fill level per second that are to be blown away.
function WindyLoads:calculateBlowRates(airSpeed, fillUnits)
--TODO: wenn das FillVolume von einer geschlossenen Abdeckung geschützt ist, dann die blowRate auf Null setzen
	local fillUnits = fillUnits or {};
	local specFillUnit = self.spec_fillUnit;
	local blowRatePerFillUnit = {};
	for fillUnitIndex, fillUnit in pairs(fillUnits) do
		local fillUnitCapacity = specFillUnit:getFillUnitCapacity(fillUnitIndex);
		if fillUnitCapacity ~= 0 then	--on some vehicles we may find fill units with a capacity of 0. This may lead to issues with division by zero
			local fillLevelPercent = specFillUnit:getFillUnitFillLevelPercentage(fillUnitIndex);
			local fillUnitFillType = specFillUnit:getFillUnitFillType(fillUnitIndex);

			if airSpeed < WindyLoads.MIN_SPEED_TO_BLOW or fillUnitFillType == nil then
			--nothing will be blown away below the global minimum speed
				blowRatePerFillUnit[fillUnitIndex] = 0;
			elseif WindyLoads.fillTypesIndexToName[fillUnitFillType] == nil or fillLevelPercent == 0 then
				blowRatePerFillUnit[fillUnitIndex] = 0;
			elseif self.spec_cover and not specFillUnit:getFillUnitSupportsToolType(fillUnitIndex, ToolType.TRIGGER) then
			--this happens when there is a cover that is closed. We don't want anything to be blown away then.
				blowRatePerFillUnit[fillUnitIndex] = 0;
			else
			--now we have to calculate for each fillUnit how fast the contents are blown away
				local blowRate = 0;
				-- we want the blow effect to be dependent on the fill level in a cubic manner.
				local blowVolumeFactor = fillLevelPercent ^ 3;
				local blowFactor = WindyLoads.blowFactors[WindyLoads.fillTypesIndexToName[fillUnitFillType]] or WindyLoads.STANDARD_BLOW_FACTOR;
				local fillWeight = WindyLoads.applicableFillTypes[WindyLoads.fillTypesIndexToName[fillUnitFillType]].massPerLiter or WindyLoads.STANDARD_MASS_PER_LITER;
				--we want the blow rate to depend on the air speed in a quadratic manner
				blowRate = blowFactor / fillWeight /100000000 * (airSpeed - WindyLoads.MIN_SPEED_TO_BLOW)^2;
				blowRatePerFillUnit[fillUnitIndex] = blowVolumeFactor * blowRate * WindyLoads.BLOW_FACTOR_GLOBAL;
			end;
		end;
	end;
	return blowRatePerFillUnit;
end;

--this function applies the effects of this mod to all applicable fillUnits of the vehicle
function WindyLoads:affectFillUnits(dt, sendNoEvent)
	local spec = self.spec_windyLoads;
	local specFillUnit = self.spec_fillUnit;
	if not spec.modInitialized or spec.vehicle == nil or spec.vehicleSupported == false then
		--there's nothing here for us to compute
		return;
	end;
	local blowRatePerFillUnit = WindyLoads.calculateBlowRates(self, spec.lastVehicleSpeed, spec.supportedFillUnits);
	spec.blowRatePerFillUnit = blowRatePerFillUnit; -- store the values
    for fillUnitIndex, fillUnit in pairs(spec.supportedFillUnits) do
		if blowRatePerFillUnit[fillUnitIndex] ~= nil and (blowRatePerFillUnit[fillUnitIndex] > 0 or spec.lastBlownVolume[fillUnitIndex] > 0) then
		--we only need to do the calculation if there is something to substract
			local fillLevelPercent = specFillUnit:getFillUnitFillLevelPercentage(fillUnitIndex);
			local fillUnitFillType = specFillUnit:getFillUnitFillType(fillUnitIndex);
			if fillLevelPercent and fillUnitFillType and spec.blowRatePerFillUnit[fillUnitIndex] > 0 then
			--at least we have something that we can work with
				local deltaVol = blowRatePerFillUnit[fillUnitIndex] * fillUnit.fillLevel / 100 * dt / 1000;
				spec.lastBlownVolume[fillUnitIndex] = spec.lastBlownVolume[fillUnitIndex] + deltaVol;
				spec.timeSinceLastSync[fillUnitIndex] = spec.timeSinceLastSync[fillUnitIndex] + dt;
				if spec.timeSinceLastSync[fillUnitIndex] > WindyLoads.SYNC_TIME_INTERVAL_MS then
					WindyLoads.blowFromFillUnit(self, fillUnitIndex, fillUnit.fillLevel-spec.lastBlownVolume[fillUnitIndex], false);
				end;
				--local appliedDelta = specFillUnit:addFillUnitFillLevel(self:getOwnerFarmId(), fillUnitIndex, deltaVol, specFillUnit:getFillUnitFillType(fillUnitIndex), ToolType.UNDEFINED, nil);
			end;
		end;
	end;
end;

--this function is used to actually perform the reduction of the fill unit caused by blown away loads
function WindyLoads:blowFromFillUnit(fillUnitIndex, newVol, sendNoEvent)
	local spec = self.spec_windyLoads;
	local specFillUnit = self.spec_fillUnit;
	
	--if self.isServer then
		local appliedDelta = specFillUnit:addFillUnitFillLevel(self:getOwnerFarmId(), fillUnitIndex, newVol - specFillUnit:getFillUnitFillLevel(fillUnitIndex), specFillUnit:getFillUnitFillType(fillUnitIndex), ToolType.UNDEFINED, nil);
		--print(string.format("appliedDelta = %.5f; toolTypeSupported = %s", appliedDelta, tostring(specFillUnit:getFillUnitSupportsToolTypeAndFillType(fillUnitIndex, ToolType.UNDEFINED, specFillUnit:getFillUnitFillType(fillUnitIndex)))));
		self:updateMass();
		spec.lastBlownVolume[fillUnitIndex] = 0;
		spec.timeSinceLastSync[fillUnitIndex] = 0;
	--end;

	if not sendNoEvent then
		WindyLoadsBlowEvent.sendEvent(self, fillUnitIndex, newVol, sendNoEvent);
	end;
end;

-- this function shall handle the necessary action if we deal with particle systems
function WindyLoads:handleParticles(dt)
	local spec = self.spec_windyLoads;	--the general spec of this mode
	if not WindyLoads.ENABLE_PARTICLES or not spec.windyParticleSpec then	--we have no valid particle system. Aborting!
		return;
	end;
	local specFillUnit = self.spec_fillUnit;	--the spec for all the fill units in the vehicle
	if not spec.supportedFillUnits or not(#(spec.supportedFillUnits) > 0) then
		if WindyLoads.DEBUG_MODE then
			print("[WindyLoads]: Cannot find supported fill units required for particle handling on vehicle '" .. spec.vehicle:getFullName() .. WindyLoads.getVehicleIDStringDebug(self) .. "' for fill unit " .. fillUnitIndex .. "!");
		end;
		return;
	end;
	for fillUnitIndex, _ in pairs(spec.supportedFillUnits) do	--now we loop through all the fill units that are applicable for this mod
		local windyParticleSpec = spec.windyParticleSpec[fillUnitIndex]; -- the spec containing all the info about the partice systems
		if windyParticleSpec ~= nil then
			if not windyParticleSpec.particlesInitialized or windyParticleSpec.psEffects == nil or windyParticleSpec.effectEmitterNode == nil then
				windyParticleSpec.particlesInitialized = false;	--remember that we found some missing particle configuration
				if WindyLoads.DEBUG_MODE then
					print("[WindyLoads]: Cannot handle particles on vehicle '" .. spec.vehicle:getFullName() .. WindyLoads.getVehicleIDStringDebug(self) .. "' for fill unit " .. fillUnitIndex .. "!");
				end;
				--we cannot do particle calculation if there's no particle system stored for the current vehicle
			else
				local currEffect = windyParticleSpec.psEffects;	--get the stored particle system for the current fill unit
				if currEffect ~= nil and spec.blowRatePerFillUnit[fillUnitIndex] ~= nil then
					local currBlowRate = spec.blowRatePerFillUnit[fillUnitIndex] or 0;	--fetch the current blow rate for determining the intensity of the particle effect
					if currBlowRate == 0 then	--nothing is blown from the fill unit
						WindyLoads:toggleParticleEffect(currEffect, false, 0);	-- turn off the particles if nothing is blown from the fill unit
					elseif currBlowRate > 0 then	--we have something that is blown from the fill unit
						local intensity = WindyLoads:getParticleIntensity(currBlowRate) or 0;	--calculate the particle intensity from the current blow rate
						WindyLoads:toggleParticleEffect(currEffect, true, intensity);	--set the particle effect active with an intensity according to the strength of the current blow rate from the respective fill unit
					end;
				end;
			end;
		end;		
	end;
end;

function WindyLoads:getParticleIntensity(blowRate)
	local result = blowRate * 40 * WindyLoads.PARTICLE_INTENSITY_FACTOR;
	result = math.max(0,math.min(result, WindyLoads.PARTICLE_MAX_INTENSITY));	--the intensity shall be in a range from zero to the maximim limit
	return result;
end;

function WindyLoads:onUpdate(dt, isActiveForInput, isActiveForInputIngnoreSelection, isSelected)
	local spec = self.spec_windyLoads;
	if not spec.modInitialized or spec.vehicle == nil or spec.vehicleSupported == false then
		--there's nothing here for us to compute
		return;
	elseif spec.vehicleSupported == nil then
	-- the vehicle should be checked, whether it is applicable for this mod
		spec.vehicleSupported, spec.supportedFillUnits = self:getVehicleIsSupportedByWindyLoads(self);
		if spec.vehicleSupported == false then
			print(string.format("[WindyLoads]: Vehicle '%s' not supported by WindyLoads mod. It will not be considered any further.", Vehicle.getFullName(self) .. WindyLoads.getVehicleIDStringDebug(self)));
			--we cannot remove all event listeners because maybe the vehicle configuration is changed to a supported one. Otherwise we wouldn't receive update events anymore
			--WindyLoads.removeFromEventListeners(g_vehicleTypeManager:getTypeByName(self.typeName));
			return;
		end;
	end;
	--local specCover = self.spec_cover;
	spec.lastVehicleSpeed = WindyLoads.getVehicleVelocity(self);
	WindyLoads:calculateCurrentRelativeAirSpeed(self);
	WindyLoads.affectFillUnits(self, dt, false);	--we handle all the effects on the fill volume itself
	if WindyLoads.ENABLE_PARTICLES then
		WindyLoads.handleParticles(self, dt);	-- now we handle the visual effects that show the blowing effect to the user
	end;
end;

function WindyLoads:onDraw()
	local spec = self.spec_windyLoads;
	
	if not spec.modInitialized or not WindyLoads.DEBUG_MODE or not spec.vehicleSupported then
		return;
	end;
	--print("windyLoads Draw called");
	if g_currentMission.controlledVehicle ~= nil then
		--only draw debug HUD when inside vehicle
		--local vehicleVel;
		local hudLineLevel = WindyLoads.hudData.hudTextPosY;
		local hudText = "Test String";
		--hudText = string.format("vehicle name: %s", g_currentMission.controlledVehicle:getRootVehicle().name);
		setTextColor(0, 1, 0, 1);
		setTextAlignment(RenderText.ALIGN_RIGHT);
		setTextVerticalAlignment(RenderText.VERTICAL_ALIGN_TOP);
		setTextBold(true);
		hudText = string.format("vehicle: v=%.1f km/h@ X=%.3f; Z=%.3f; Y=%.3f; Wind: v=%.3f km/h@ X=%.3f; Z=%.3f",
			spec.lastVehicleSpeed,
			spec.debugInfo.vehicleWorldDirectionX,
			spec.debugInfo.vehicleWorldDirectionZ,
			spec.debugInfo.vehicleWorldDirectionY,
			spec.debugInfo.windVelocity,
			spec.debugInfo.windDirX,
			spec.debugInfo.windDirZ);
			
		
		--renderText(spec.hudData.hudTextPosX, spec.hudData.hudTextPosY, spec.hudData.hudTextSize, "Test Text");
		renderText(WindyLoads.hudData.hudTextPosX, hudLineLevel, WindyLoads.hudData.hudTextSize, hudText);
		
		hudLineLevel = hudLineLevel - WindyLoads.hudData.hudTextSize*1.1;
		hudText = string.format("airSpeed = %.4f", spec.lastAirSpeed);
		renderText(WindyLoads.hudData.hudTextPosX, hudLineLevel, WindyLoads.hudData.hudTextSize, hudText);
		
		for fillUnitIndex, fillUnit in pairs(spec.supportedFillUnits) do
			hudLineLevel = hudLineLevel - WindyLoads.hudData.hudTextSize*1.1;
			hudText = string.format("fillUnit [%s]: fillLevel = %.2f; blowRate = %.4f", tostring(fillUnitIndex), fillUnit.fillLevel, spec.blowRatePerFillUnit[fillUnitIndex] or 0);
			if WindyLoads.ENABLE_PARTICLES then
                hudText = hudText .. "; partIntens: " .. string.format("%.4f", WindyLoads:getParticleIntensity(spec.blowRatePerFillUnit[fillUnitIndex] or 0));
            end;
            renderText(WindyLoads.hudData.hudTextPosX, hudLineLevel, WindyLoads.hudData.hudTextSize, hudText);
		end;
		
		--print("windyLoads HUD Text rendered");
		setTextColor(1,1,1,1);
		setTextAlignment(RenderText.ALIGN_LEFT);
		setTextVerticalAlignment(RenderText.VERTICAL_ALIGN_BASELINE);
		setTextBold(false);
		
	end;
end;

--this function shall set the state of the particle effect according to the given values
function WindyLoads:toggleParticleEffect(effectNode, state, intensity)
	--local effectNode = effect;
	if effectNode.isEmitting ~= state then
		ParticleUtil.setEmittingState(effectNode, state);
	end;
	if state == true and intensity ~= nil and effectNode ~= nil then
		WindyLoads:setParticleIntensity(effectNode, intensity);
	end;
end;

--this function shall set the particle intensity based on the given value. It will represent the intensity of the blowing effect from the fill unit
function WindyLoads:setParticleIntensity(effectNode, intensity)
	if effectNode == nil then
		return;
	end;
	if getHasShaderParameter(effectNode.shape, "colorAlpha") then
		setShaderParameter(effectNode.shape, "colorAlpha", 200/255, 179/255, 122/255, intensity, false); --the color values create a nice generally acceptable smoke color
	else
		if WindyLoads.DEBUG_MODE then
			print("[WindyLoads]: Missing shader attribute 'colorAlpha' in WindyLoads particle system material.");
		end;
		Logging.error("Missing shader attribute 'colorAlpha' in WindyLoads particle system material");
	end;
end;
